/******************************************************************
*******************************************************************
This is the header file for the code files of the LN project. It defines the
classes used, data structures, as well as some small functions. The code 
files are four: LNutil.cpp, that contains the functions related to 
manipulation of data structures ( eg create and delete cells, etc). 
LNio.cpp contains functions to read and save files. LNmain.cpp defines the main routine.

The main data structures of the project are as follows: The matrix M in 
the class LymphNode is a matrix of 400 x 25 microcompartments. Each 
microcompartment has two pointers to cells (any kind). This matrix has 
pointers to cells, no actual cells.

To store cells there are three vectors in the class LymphNode. LH 
(helper Tcells), LK (killer Tcells) and LD (dendritic cells) . Each vector 
is actually a "sparse matrix": each element contains 
a pair of coordinates (x, y) and a pointer to a cell. A sparse matrix is 
used in computer science when a matrix has a lot of "zeros" , in this case 
empty sites. (about 1000 cells or less vs 10000 sites in the matrix)
The size of each vector changes dynamically, this reflects the fact that 
cells appear and dissapear in the Lymphnode. Every time a new cell is 
added, the cell is created in memory and added to the corresponding 
vector. Every time a cells dies or goes down the efferent sites, the cell 
is deleted from the vector and its memory released. The fact that each 
element in a vector contains a cell pointer, and not the actual cell, 
ensures that since a cell is created to the moment it disappears its 
position in memory remains static. What actually changes is its 
coordinates stored in the vector. 

The three vectors contain all information needed, so the matrix M is 
actually redundant. The only purpose of having the matrix M is to speed 
up when a cell has to look at its neighbors. If the matrix M was not 
there, the cell can look at what cells are its neighbors by browsing the 
vectors, but this would be much slower. In summary, The matrix M and cell 
vectors create a good compromise between memory used and time. 

*******************************************************************
*******************************************************************/

#ifndef INC_LN_H
#define INC_LN_H

#include <vector>
#include <iostream>
#include <fstream>
#include <time.h>
#include <stdio.h>
#include "cran32.h"
#include "LNpar.h"

using namespace std;

const int MROWS = 25; // height
const int MCOLS = 200; // width
const int MDCELLS = 200; // for hContacts and kContacts arrays

// const int MROWS = 100; // height
// const int MCOLS = 100; // width

void checkRowsAndCols(int row, int col, string msg);
bool withinRowsAndCols(int row, int col);

// Helper T cell states
enum h_states {H_RESTING = 1, H_NOTMOVE_BY_DC, H_ACTIVE, H_EFFECTOR, H_BOUND_TO_IDC, H_DEAD};

// Killer T cell states
enum k_states { K_RESTING = 1, K_NOTMOVE_BY_LDC, K_ACTIVE, K_EFFECTOR, K_MEMORY, K_DEAD};

// Dendritic cell states
enum d_states { D_IMMATURE = 0, D_MATURE = 1, D_LICENSED, D_DEAD};

enum DIRECTIONS {NONE, EAST, SOUTH, WEST, NORTH};

// Class definitions

// The class cell is an abstract class. It is the mother class of each 
// type of cell. Sort of a "stem cell"
class Cell 
{
 public:
  int id; // unique identifier
  long birthtime; // birthtime of the cell in its current state
  int state; // state of the cell
  int cognate;
  int next_state; // this is needed to update synchronously
  int time_in; // time the cell appeared on the grid

 protected:
  bool isDcell;

 public:
  // Constructor. Initializes all fields
  Cell() {
    id = 0;
    birthtime = 0;
    state = 0;
    next_state = 0; // added LAB June 5
    time_in = 0; // added LAB June 5
    cognate = 0;
  }
 
  // These three functions help to identify the kind of cell by the id 
  // number 
  inline int is_hcell() {
    return id % 3 == 2;
  }
  
  inline int is_kcell() {
    return id % 3 == 1;
  }
  
  inline int is_tcell() {
    return !isDcell;
  }
  
  inline int is_dcell() {
    return id % 3 == 0;
  }
};

// Class Tcell: its the mother of helper and killer clases. It is the 
// daughter of class Cell
class Tcell : public Cell
{
 public:
  bool inFirstThousandTCells; // whether to keep track of pathLength and distance to HEV
  DIRECTIONS persistentPath; // for calculating pathLengths
  int persistentPathLength; // for calculating pathLengths
  int birthRow; // for determining lifetimeDisplacement
  int birthCol; // for determining lifetimeDisplacement
  int division_counter; // to control the division when activated
  int bound_to_dcell; // The id of a Dendritic cell that this cell is bound to
  int direction; // The direction the cell went last time so it won't go back on itself. can be combined with persistentPath
  int not_move; // time interval that the cell is bound to a Dcell
  long time_bound; // time that the cell bounded to a dcell
  float contact_history; 
  float binding_history;
  int n_DcContacts;
  int n_uniqueDcContacts;
  int searchTime; // time till contacting first MDC
  float binding_prob;
  int bound; // 0(no) or 1(yes)
  float activation_prob;
  int activated; // 0(no) or 1(yes)
  int activation_time; // time since it appeared on the grid
  int time_activated; // time when tcell got activated 
  int time_became_effector; // time when tcell became effector
  int transitTime; // time spent on the grid
  vector<int> cognateDCContacts;
  vector<int> dcContacted;
  vector<int> dcContactedActive;

  // Constructor. It calls the constructor of the mother class
  Tcell(): Cell() {
    isDcell = false;
    inFirstThousandTCells = false;
    division_counter = 0;
    bound_to_dcell = 0;
    not_move = 12 * 600; // tcell stays bound to dcell to max 12 hours 
    direction = 0;
    contact_history = 0;
    binding_history = 0;
    n_DcContacts = 0;
    n_uniqueDcContacts = 0;
    searchTime = -1;
    binding_prob = 0;
    bound = 0;
    activation_prob = 0;
    activated = 0;
    activation_time = 0;
    time_activated = 0;
    time_became_effector = 0;
    transitTime = -1;
  }

  double calcDist(int row, int col);
  void add_dcell(long dcell_id);
  void remove_dcell(long dcell_id);
  void add_dcellActive(long dcell_id);
  void remove_dcellActive(long dcell_id);
};

// Class for helper cells
class h_cell : public Tcell 
{
 public:
  h_cell(): Tcell(){
  }
};

// Class for killer cells 
class k_cell : public Tcell 
{
 public:
  k_cell(): Tcell(){
  }
};

// Dendritic cell class 
class d_cell : public Cell 
{
 public:
  vector<int> tbind; // a vector of Tcell id's , that are bound to the cell
  vector<int> kcells;
  vector<int> hcells;
  
  double initial_pMHC; 
  double n_pMHC; // number of antignes on the DC, value needs to be double to account for decay
  int licensing_time; // time when MDC becomes licensed
  
  int n_hcell_contacts;
  int n_unique_hcell_contacts;
  
  int n_kcell_contacts;
  int n_unique_kcell_contacts;
  
  int n_hcell_contactsActive;
  int n_unique_hcell_contactsActive;
  
  int n_kcell_contactsActive;
  int n_unique_kcell_contactsActive;
  
  // Constructor
  d_cell(): Cell(){
	n_pMHC = 0;
	state = D_MATURE;
    isDcell = true;
    licensing_time = 0;
    
    n_hcell_contacts = 0;
    n_unique_hcell_contacts = 0;
    
    n_kcell_contacts = 0;
    n_unique_kcell_contacts = 0;
    
    n_hcell_contactsActive = 0;
    n_unique_hcell_contactsActive = 0;
    
    n_kcell_contactsActive = 0;
    n_unique_kcell_contactsActive = 0;
  }
  
  // These two functions add and remove a Tcell from the tbind vector

  void add_tcell(long hcell_id);
  
  void remove_tcell(long hcell_id);
  void add_hcell(long hcell_id);
  
  void remove_hcell(long hcell_id);
  
  void add_kcell(long hcell_id);
  
  void remove_kcell(long hcell_id);
  // function to generate random numbers with normal distribution
  double normal(const double & , const double & );
  
};

// microcompartment is the base class for the matrix. It contains a 
// cell pointer
class microcompartment
{
 public:
  static const int NUM_CELLS = 4;
  Cell* cell[NUM_CELLS];
  
  // check whether there is space for a tcell
  // not the opposite of is_full or is_empty
  inline bool free_site() {
    if(cell[0] == NULL){
      return true;
    }
    if(cell[0]->is_dcell()){
      return false;
    }
    return !is_full();
  }

  inline bool is_full() {
    for(int i = 0; i < NUM_CELLS; i++){
      if(cell[i] == NULL){
	return false;
      }
    }
    return true;
  }
  
  // function to know if there is a dcell in the microcompartment
  inline bool has_dcell() {
    return (cell[0] && cell[0]->is_dcell());
  }   
  
  // technically space 0 should be the only celll that is needed to determine whether the whole microcompartment is full or not, but I am having errors, so I am going to check every cell.  Otherwise, a cell in the 2nd or whatever slot gets over-written
  inline bool is_empty() {
    if(cell[0] != NULL && cell[0]->is_dcell()){
      return false;
    }
    for(int i = 0; i < NUM_CELLS; i++){
      if(cell[i] != NULL){
	return false;
      }
    }
    return true;
  }

  inline int which_site(){
    if(cell[0] == NULL){
      return 0;
    }
    for(int i = 0; i < NUM_CELLS; i++){
      if(cell[i] == NULL){
	return i;
      }
    }
    for(int i = 0; i < NUM_CELLS; i++){
      cerr << "cell[" << i << "]: " << cell[i] << endl;
    }
    cerr << "There is no free site.  Exiting." << endl;
    exit(5);
  }

  // the index of the first available site is the number of actual Cell's in a microcompartment
  inline int how_many_cells(){
    for(int i = 0; i < NUM_CELLS; i++){
      if(cell[i] == NULL){
	return i;
      }
    }
    return NUM_CELLS;
  }

  // shifts cells to fill a space after a cell has been removed
  inline void shift_cells(const int& changedCell){
    cell[changedCell] = NULL;
    for(int i = changedCell; i < NUM_CELLS - 1; i++){
      cell[i] = cell[i + 1];
    }
    cell[NUM_CELLS - 1] = NULL; 
  }

  // returns the position of a referenced cell, quits if its not there
  inline int find_cell(const Cell* ptr){
    for(int i = 0; i < NUM_CELLS; i++){
      if(cell[i] == ptr){
	return i;
      }
    }
    return -1;
    cerr << "The cell is not at this site.  Exiting." << endl;
    exit(26);
  }
};

// Class to define an element in the cell vectors. It has a pair of (x,y) 
// coordinates and a cell pointer
class cellpos 
{
 public:
  int row;
  int col;
  Cell* cell_pointer;
  
  void set_position(int r, int c){
    row = r;
    col = c;
  }
};

// This the class to define a vector of Cellpos. 
class cell_vector 
{
  public : 
    vector<cellpos> v;
  
  // This function adds a cellpos to the vector
  void push_back(cellpos cp) {
    v.push_back(cp);
  }
  
  // This function returns the position of a cell in the vector
  int whereis(Cell* ptr) {
    for(int i = v.size() - 1; i >= 0; i--){
      if (ptr == v[i].cell_pointer){ 
	return i;
      }
    }
    cerr << "Error in whereis.  ptr is not in vector.  Exiting." << endl;
    exit(12);
  }

  int whereis(int id){
    for(int i = v.size() - 1; i >= 0; i--){
      if(v[i].cell_pointer->id == id){
	return i;
      }
    }
    cerr << "Warning:  Error in whereis.  integer value: " << id << " is not in vector." << endl;
    return -1;
    // exit(57);
  }
 
  void freeIntVector(vector<int>& vi) {
    vi.clear();
  }
  
  // This function releases the memory used by an EMPTY vector of doubles
  void freeDoubleVector(vector<double>& vi) {
    vi.clear();
  }

  // This funcion deletes a cellpos from the vector
  void erase(Cell *ptr) {
    erase(elem(whereis(ptr)));
  }

  void erase(cellpos *cp) {
    if (!(cp->cell_pointer->is_dcell())) // hcell or kcell ptr
    {  
	if (cp->cell_pointer->is_hcell())
	{     
      		((Tcell *)cp->cell_pointer)->dcContacted.clear();
      		freeIntVector( ((h_cell *)cp->cell_pointer)->dcContacted );
      
      		((Tcell *)cp->cell_pointer)->dcContactedActive.clear();
      		freeIntVector( ((h_cell *)cp->cell_pointer)->dcContactedActive );
    	}  
    	else if (cp->cell_pointer->is_kcell()) {
      
      		((k_cell *)cp->cell_pointer)->dcContacted.clear();
      		freeIntVector( ((k_cell *)cp->cell_pointer)->dcContacted );
      
      		((k_cell *)cp->cell_pointer)->dcContactedActive.clear();
      		freeIntVector( ((k_cell *)cp->cell_pointer)->dcContactedActive );
      
    	}
    }
    else {  // dcell ptr
      ((d_cell *)cp->cell_pointer)->tbind.clear();
      freeIntVector( ((d_cell *)cp->cell_pointer)->tbind );
      ((d_cell *)cp->cell_pointer)->hcells.clear();
      freeIntVector( ((d_cell *)cp->cell_pointer)->hcells );
      ((d_cell *)cp->cell_pointer)->kcells.clear();
      freeIntVector( ((d_cell *)cp->cell_pointer)->kcells );
    }
    delete cp->cell_pointer; // release memory
    v.erase((vector <cellpos>::iterator)cp); 
 }

  // This function returns the vector size
  int size() {
    return v.size();
  }
 
  // returns the element at position i 
  inline cellpos* elem(int i) {
    // when the program is working, comment out these bounds checking
    /*if(i >= v.size()){
      cerr << i << " is beyond the bounds of this vector.  Exiting. " << endl;
      exit(27);
    }
    if(i < 0){
      cerr << i << " is not a positive integer, so there is no vector element there.  Exiting. " << endl;
      exit(28);
      }*/
    return &v[i];
  }

  void clear() {
    vector<cellpos>::iterator iter;
    for (int i = v.size() - 1; i >= 0; i--) {
      iter = (vector <cellpos>::iterator)&v[i];
      delete v[i].cell_pointer; // release memory
      v.erase(iter);
    }
  }
};

// This is the main class of the program. It defines a LymphNode. It 
// contains the matrix M, the vectors LH, LK and LD, as well as all the 
// parameters for the simulations. The implementation of the functions for 
// this class are in the 4 code files.
class LymphNode {
 public:
  Par p;
  vector<int> lifetimeDisplacement;
  int allFirstThousandTCells;
  // parameters to control the update intervals for each kind of cell
  int brickWallCounter;
  int step_tcell;
  int step_dcell; 
    
  int histo[960];
  int DC_space;
  float pMHC_decay;
  vector< vector<int> > hMDCContactsTotal;
  vector<int> hMDCContactsNaive;
  vector<int> hMDCContactsNaiveCognate;
  vector<int> hMDCContactsUnique;
  vector<int> hMDCContactsActiveUnique;
  vector<int> hMDCContactsCognate;
  vector<int> hMDCContactsNonCognate;
  vector<int> hMDCBoundTotal;
  vector<int> hMDCActivatedTotal;

  vector< vector<int> > hLDCContactsTotal;
  vector<int> hLDCContactsNaive;
  vector<int> hLDCContactsNaiveCognate;
  vector<int> hLDCContactsUnique;
  vector<int> hLDCContactsActiveUnique;
  vector<int> hLDCContactsCognate;
  vector<int> hLDCContactsNonCognate;
  vector<int> hLDCBoundTotal;
  vector<int> hLDCActivatedTotal;
    
  vector< vector<int> > kMDCContactsTotal;
  vector<int> kMDCContactsNaive;
  vector<int> kMDCContactsNaiveCognate;
  vector<int> kMDCContactsUnique;
  vector<int> kMDCContactsActiveUnique;
  vector<int> kMDCContactsCognate;
  vector<int> kMDCContactsNonCognate;
  vector<int> kMDCBoundTotal;
  vector<int> kMDCActivatedTotal;
  
  vector< vector<int> > kLDCContactsTotal;
  vector<int> kLDCContactsNaive;
  vector<int> kLDCContactsNaiveCognate;
  vector<int> kLDCContactsUnique;
  vector<int> kLDCContactsActiveUnique;
  vector<int> kLDCContactsCognate;
  vector<int> kLDCContactsNonCognate;
  vector<int> kLDCBoundTotal;
  vector<int> kLDCActivatedTotal;
  
  vector< vector<float> > hCognate; // store congnate resting hcell id, activation_prob, activation_time
  vector< vector<float> > kCognate;

  vector<int> pathLengths;
  int searchTime[1000];
  
  // the number of sites for efferent, t sources and afferent sites
  int NSOURCES, NEFFERENT, NAFFERENT, NIDC_AFFERENT;
  
  // These are counters to monitor the events in the simulation. 
  int n_t4_cells_leave; // # naive t4 cells leave
  int n_t8_cells_leave; // # naive t8 cells leave
  
  int n_t4a_cells_leave; // # active t4 cells leave
  
  int n_t4e_cells_leave; // # effector t4 cells leave
  
  int n_t8a_cells_leave; // # active t8 cells leave
  
  int n_t8m_cells_leave; // # memory t8 cells leave
  
  int n_t8e_cells_leave; // # effector t8 cells leave
  
  int NRH,NAH,NEH,NRK,NAK,NEK,NMK,NID,NMD,NLD;
  int n_hcells;  // number of hcells created
  int n_kcells;  // number of kcells created
  int n_cog_hcells;
  int n_cog_kcells;
  // new counters for dc-tcell interaction
  int n_dcell_contacts;
  int n_unique_dcell_contacts;
  
  int COGNATE_0;
  int NEW_H_CELLS;
  int NEW_K_CELLS;

  bool SAVE_FRAME; // boolean parameter to save frame files or not
  bool PROG; // boolean parameter to print progress to oupput or not
  bool CHRONIC; // indicates whether to introduce new mature dcells over time or not
  CRan32 rnd; // Random number generator
  long counter; // main counter (time)
  
  int* efferent; // array that contains efferent coordinates
  int* afferent; // array that contains afferent coordinates
  int* idc_afferent;
  int* sources; // array that contains T sources coordinates
  
  ofstream statfile;
  ofstream pathLengthFile;
  ofstream pMHCfile;
  ofstream hCogContactsFile;
  ofstream kCogContactsFile;
  ofstream DC_recruitmentFile;
  ofstream Tcells_createdFile;
    
  char filename[100]; // just a string to store filenames
  int synch; // Type of update : 0 synchronous, 1,2 asynchronous
  vector<int> firstThousandHCells;
  microcompartment M[MROWS][MCOLS]; // Matrix
  cell_vector LD; // Cellpos vectors
  cell_vector LH;
  cell_vector LK;
  
  long index_t4; // index for helper cell id's
  long index_t8; // index for killer cell id's
  long index_dc; // index for dendritic cell id's
  
  // Constructor 
  LymphNode():rnd(time(NULL),time(NULL)+99999) {
    readParFile("LN.par");
    //cout <<"Read the par file" << endl;
    allFirstThousandTCells = 1000;
    pathLengths.resize(1000, 0);
    lifetimeDisplacement.resize(1000, 0);

    DC_space= 25 * p.PEAK_DEN_CELLS();
    //int DC_contacts =  DC_space;
    vector<int> temp(4);
    
    hMDCContactsTotal.resize(DC_space, temp);
    hMDCContactsNaive.resize(DC_space);
    hMDCContactsNaiveCognate.resize(DC_space);
    hMDCContactsUnique.resize(DC_space);
    hMDCContactsActiveUnique.resize(DC_space);
    hMDCContactsCognate.resize(DC_space);
    hMDCContactsNonCognate.resize(DC_space);
    hMDCBoundTotal.resize(DC_space);
    hMDCActivatedTotal.resize(DC_space);
    
    hLDCContactsTotal.resize(DC_space, temp);
    hLDCContactsNaive.resize(DC_space);
    hLDCContactsNaiveCognate.resize(DC_space);
    hLDCContactsUnique.resize(DC_space);
    hLDCContactsActiveUnique.resize(DC_space);
    hLDCContactsCognate.resize(DC_space);
    hLDCContactsNonCognate.resize(DC_space);
    hLDCBoundTotal.resize(DC_space);
    hLDCActivatedTotal.resize(DC_space);
    
    kMDCContactsTotal.resize(DC_space, temp);
    kMDCContactsNaive.resize(DC_space);
    kMDCContactsNaiveCognate.resize(DC_space);
    kMDCContactsUnique.resize(DC_space);
    kMDCContactsActiveUnique.resize(DC_space);
    kMDCContactsCognate.resize(DC_space);
    kMDCContactsNonCognate.resize(DC_space);
    kMDCBoundTotal.resize(DC_space);
    kMDCActivatedTotal.resize(DC_space);
    
    kLDCContactsTotal.resize(DC_space, temp);
    kLDCContactsNaive.resize(DC_space);
    kLDCContactsNaiveCognate.resize(DC_space);
    kLDCContactsUnique.resize(DC_space);
    kLDCContactsActiveUnique.resize(DC_space);
    kLDCContactsCognate.resize(DC_space);
    kLDCContactsNonCognate.resize(DC_space);
    kLDCBoundTotal.resize(DC_space);
    kLDCActivatedTotal.resize(DC_space);
        
    time_t timer;  // Calendar time, time elapsed
    tm *timest; //  struct tm in time.h for local time (broken down time)
    synch = 0;
    
    step_tcell = 20; // 5 in other file  
    step_dcell = 80; // 20 in other file // 8 min * 60 sec/min * 1 timestep/6 sec
    
    SAVE_FRAME = false; // may be reset in arguments of main()
    PROG = true; // may be reset in arguments of main()
    CHRONIC = false; // may be reset in arguments of main()
    
    time(&timer);
    timest = localtime(&timer); 
    // sprintf is C function to write formatted data to string 
    // int sprintf ( char * str, const char * format, ... );
    sprintf(filename,"data/LN-%02d%02d%02d-%02d%02d-",timest->tm_mon+1,
	    timest->tm_mday,timest->tm_year,timest->tm_hour,timest->tm_min);
    
    index_t4 = 2; // for giving ids to cells
    index_t8 = 1; // add 3 to each new id
    index_dc = 0; // then use the % operator to distinguish them
}
  
  // These are the prototypes for all the functions of the class LymphNode
  // They are implemented in the five code files
  
  void create_hcell(int row, int col,int cell_index,int h_age, 
			       int cell_state, int division_counter, int cognate, int bCounter);
  
  void delete_hcell(int row, int col, int cell_index);
  
  
  void create_kcell(int row, int col,int cell_index, int k_age,
			       int cell_state, int division_counter, int cognate, int bCounter);

  void delete_kcell(int row, int col, int cell_index);
  
  int  create_dcell(int row, int col, int d_age, 
			        int cognate);
  
  void delete_dcell(int row, int col);
  
  void recruit_h_cells(); 
  
  void recruit_k_cells(); 
  
  void recruit_d_cells(); 
  
  void remove_dead_cells();
  
  bool freeAndExistingSite(int row, int col);
  
  void set_helpercells_random();
  
  void set_killercells_random();
  
  void set_dendriticcells_random();
  
  int next_state_helperTcell(cellpos *ptr, int nDC);
  
  int next_state_helperTcell_resting(cellpos* ptr, int nDC);

  int next_state_helperTcell_notmove(cellpos* ptr);

  int next_state_helperTcell_active(cellpos* ptr);

  int next_state_helperTcell_effector(cellpos* ptr);

  int next_state_helperTcell_bound(cellpos* ptr);

  int next_state_killerTcell(cellpos* ptr);

  int next_state_killerTcell_resting(cellpos* ptr);

  int next_state_killerTcell_notmove(cellpos* ptr);

  int next_state_killerTcell_active(cellpos* ptr);

  int next_state_killerTcell_effector(cellpos* ptr);

  int next_state_killerTcell_memory(cellpos* ptr);
  
  int next_state_dcell(cellpos* ptr);

  void find_most_probable_row_col(const cellpos* cpos, 
					     int & newrow, int & newcol);
  
  void move_helpercell();
  
  void move_killercell();
  
  void move_dendriticcell();
  
  void move_dendriticcell_loop(cellpos* mac, d_cell* pd);

  void move_dendriticcell_hlpr(const int& dcellID, int row, int col, 
					  int newRow, int newCol);

  void save_data(ofstream &stream, char *data,int len);

  int readIniFile(char *filename);
  
  int readParFile(char *filename);
  
  void save_frame();
  
  void save_frame_bin();

  void initializeValues();

  void reinitializeValues();

  int recruitDCs(int t, double a, double b, double c, double d);
 
  int run();
  
  void update(int synch, int nDC);
  
  void freeLNMemory();
};

#endif 
